@@ -34,6 +34,7 @@
 extern "C" {
 #endif /* __cplusplus */
 
+#include "config.h"
 #include "fping.h"
 #include "options.h"
 #include "optparse.h"
@@ -44,7 +45,6 @@ extern "C" {
 #include <stdio.h>
 #include <time.h>
 
-#include "config.h"
 #include "seqmap.h"
 
 #ifdef HAVE_UNISTD_H
@@ -80,6 +80,17 @@ extern "C" {
 
 #include <sys/select.h>
 
+/*** compatibility ***/
+
+/* Mac OS X's getaddrinfo() does not fail if we use an invalid combination,
+ * e.g. AF_INET6 with "127.0.0.1". If we pass AI_UNUSABLE to flags, it behaves
+ * like other platforms. But AI_UNUSABLE isn't available on other platforms,
+ * and we can safely use 0 for flags instead.
+ */
+#ifndef AI_UNUSABLE
+#define AI_UNUSABLE 0
+#endif
+
 /*** externals ***/
 
 extern char* optarg;
@@ -230,15 +241,19 @@ HOST_ENTRY* ev_first;
 HOST_ENTRY* ev_last;
 
 char* prog;
-int ident; /* our pid */
+int ident4 = 0; /* our icmp identity field */
 int socket4 = -1;
+int using_sock_dgram4 = 0;
 #ifndef IPV6
 int hints_ai_family = AF_INET;
 #else
+int ident6 = 0;
 int socket6 = -1;
 int hints_ai_family = AF_UNSPEC;
 #endif
 
+volatile sig_atomic_t status_snapshot = 0;
+
 unsigned int debugging = 0;
 
 /* times get *100 because all times are calculated in 10 usec units, not ms */
@@ -280,7 +295,6 @@ struct timeval start_time;
 struct timeval end_time;
 struct timeval last_send_time; /* time last ping was sent */
 struct timeval next_report_time; /* time next -Q report is expected */
-struct timezone tz;
 
 /* switches */
 int generate_flag = 0; /* flag for IP list generation */
@@ -317,6 +331,7 @@ void print_per_system_splits(void);
 void print_netdata(void);
 void print_global_stats(void);
 void main_loop();
+void sigstatus();
 void finish();
 char* sprint_tm(int t);
 void ev_enqueue(HOST_ENTRY* h);
@@ -359,9 +374,9 @@ int main(int argc, char** argv)
         usage(0);
     }
 
-    socket4 = open_ping_socket_ipv4(ping_data_size);
+    socket4 = open_ping_socket_ipv4(&using_sock_dgram4);
 #ifdef IPV6
-    socket6 = open_ping_socket_ipv6(ping_data_size);
+    socket6 = open_ping_socket_ipv6();
     /* if called (sym-linked) via 'fping6', imply '-6'
      * for backward compatibility */
     if (strstr(prog, "fping6")) {
@@ -369,6 +384,11 @@ int main(int argc, char** argv)
     }
 #endif
 
+    memset(&src_addr, 0, sizeof(src_addr));
+#ifdef IPV6
+    memset(&src_addr6, 0, sizeof(src_addr6));
+#endif
+
     if ((uid = getuid())) {
         /* drop privileges */
         if (setuid(getuid()) == -1)
@@ -376,7 +396,7 @@ int main(int argc, char** argv)
     }
 
     optparse_init(&optparse_state, argv);
-    ident = getpid() & 0xFFFF;
+    ident4 = ident6 = getpid() & 0xFFFF;
     verbose_flag = 1;
     backoff_flag = 1;
     opterr = 1;
@@ -427,15 +447,17 @@ int main(int argc, char** argv)
     while ((c = optparse_long(&optparse_state, longopts, NULL)) != EOF) {
         switch (c) {
         case '4':
+#ifdef IPV6
             if (hints_ai_family != AF_UNSPEC) {
                 fprintf(stderr, "%s: can't specify both -4 and -6\n", prog);
                 exit(1);
             }
             hints_ai_family = AF_INET;
+#endif
             break;
         case '6':
 #ifdef IPV6
-            if (hints_ai_family != AF_UNSPEC) {
+            if (hints_ai_family != AF_UNSPEC && hints_ai_family != AF_INET6) {
                 fprintf(stderr, "%s: can't specify both -4 and -6\n", prog);
                 exit(1);
             }
@@ -607,7 +629,7 @@ int main(int argc, char** argv)
             break;
 
         case 'H':
-            if (!(ttl = (u_int)atoi(optparse_state.optarg)))
+            if (!(ttl = (unsigned int)atoi(optparse_state.optarg)))
                 usage(1);
             break;
 
@@ -626,7 +648,7 @@ int main(int argc, char** argv)
 
         case 'x':
             if (!(min_reachable = (unsigned int)atoi(optparse_state.optarg)))
-                usage(1); 
+                usage(1);
             break;
 
         case 'f':
@@ -658,12 +680,14 @@ int main(int argc, char** argv)
             if (socket4 >= 0) {
                 if (setsockopt(socket4, SOL_SOCKET, SO_BINDTODEVICE, optparse_state.optarg, strlen(optparse_state.optarg))) {
                     perror("binding to specific interface (SO_BINTODEVICE)");
+                    exit(1);
                 }
             }
 #ifdef IPV6
             if (socket6 >= 0) {
                 if (setsockopt(socket6, SOL_SOCKET, SO_BINDTODEVICE, optparse_state.optarg, strlen(optparse_state.optarg))) {
                     perror("binding to specific interface (SO_BINTODEVICE), IPV6");
+                    exit(1);
                 }
             }
 #endif
@@ -962,12 +986,12 @@ int main(int argc, char** argv)
         exit(num_noaddress ? 2 : 1);
     }
 
-    if (src_addr_set && socket4 >= 0) {
-        socket_set_src_addr_ipv4(socket4, &src_addr);
+    if (socket4 >= 0) {
+        socket_set_src_addr_ipv4(socket4, &src_addr, &ident4);
     }
 #ifdef IPV6
-    if (src_addr6_set && socket6 >= 0) {
-        socket_set_src_addr_ipv6(socket6, &src_addr6);
+    if (socket6 >= 0) {
+        socket_set_src_addr_ipv6(socket6, &src_addr6, &ident6);
     }
 #endif
 
@@ -1006,8 +1030,10 @@ int main(int argc, char** argv)
 #endif
 
     signal(SIGINT, finish);
+    signal(SIGQUIT, sigstatus);
+    setlinebuf(stdout);
 
-    gettimeofday(&start_time, &tz);
+    gettimeofday(&start_time, NULL);
     current_time = start_time;
 
     if (report_interval) {
@@ -1279,7 +1305,12 @@ void main_loop()
                 ; /* process other replies in the queue */
         }
 
-        gettimeofday(&current_time, &tz);
+        gettimeofday(&current_time, NULL);
+
+        if (status_snapshot) {
+            status_snapshot = 0;
+            print_per_system_splits();
+        }
 
         /* Print report */
         if (report_interval && (loop_flag || count_flag) && (timeval_diff(&current_time, &next_report_time) >= 0)) {
@@ -1294,6 +1325,26 @@ void main_loop()
     }
 }
 
+/************************************************************
+
+  Function: sigstatus
+
+*************************************************************
+
+  Inputs:  void (none)
+
+  Description:
+
+  SIGQUIT signal handler - set flag and return
+
+************************************************************/
+
+void sigstatus()
+{
+    status_snapshot = 1;
+}
+
+
 /************************************************************
 
   Function: finish
@@ -1313,7 +1364,7 @@ void finish()
     int i;
     HOST_ENTRY* h;
 
-    gettimeofday(&end_time, &tz);
+    gettimeofday(&end_time, NULL);
 
     /* tot up unreachables */
     for (i = 0; i < num_hosts; i++) {
@@ -1380,8 +1431,6 @@ void print_per_system_stats(void)
     HOST_ENTRY* h;
     int resp;
 
-    fflush(stdout);
-
     if (verbose_flag || per_recv_flag)
         fprintf(stderr, "\n");
 
@@ -1461,11 +1510,18 @@ void print_netdata(void)
     int i, avg;
     HOST_ENTRY* h;
 
-    fflush(stdout);
-
     for (i = 0; i < num_hosts; i++) {
         h = table[i];
 
+        /* if we just sent the probe and didn't receive a reply, we shouldn't count it */
+        h->discard_next_recv_i = 0;
+        if (h->waiting && timeval_diff(&current_time, &h->last_send_time) < h->timeout) {
+            if (h->num_sent_i) {
+                h->num_sent_i--;
+                h->discard_next_recv_i = 1;
+            }
+        }
+
         if (!sent_charts) {
             printf("CHART fping.%s_packets '' 'FPing Packets for host %s' packets '%s' fping.packets line 110020 %d\n", h->name, h->host, h->name, report_interval / 100000);
             printf("DIMENSION xmt sent absolute 1 1\n");
@@ -1483,15 +1539,6 @@ void print_netdata(void)
             /* printf("DIMENSION lost '' absolute 1 1\n"); */
         }
 
-        /* if we just sent the probe and didn't receive a reply, we shouldn't count it */
-        h->discard_next_recv_i = 0;
-        if (h->waiting && timeval_diff(&current_time, &h->last_send_time) < h->timeout) {
-            if (h->num_sent_i) {
-                h->num_sent_i--;
-                h->discard_next_recv_i = 1;
-            }
-        }
-
         printf("BEGIN fping.%s_quality\n", h->name);
         /*
         if( h->num_recv_i <= h->num_sent_i )
@@ -1544,12 +1591,10 @@ void print_per_system_splits(void)
     HOST_ENTRY* h;
     struct tm* curr_tm;
 
-    fflush(stdout);
-
     if (verbose_flag || per_recv_flag)
         fprintf(stderr, "\n");
 
-    gettimeofday(&current_time, &tz);
+    gettimeofday(&current_time, NULL);
     curr_tm = localtime((time_t*)&current_time.tv_sec);
     fprintf(stderr, "[%2.2d:%2.2d:%2.2d]\n", curr_tm->tm_hour,
         curr_tm->tm_min, curr_tm->tm_sec);
@@ -1609,7 +1654,6 @@ void print_per_system_splits(void)
 
 void print_global_stats(void)
 {
-    fflush(stdout);
     fprintf(stderr, "\n");
     fprintf(stderr, " %7d targets\n", num_hosts);
     fprintf(stderr, " %7d alive\n", num_alive);
@@ -1662,7 +1706,7 @@ int send_ping(HOST_ENTRY* h)
     int myseq;
     int ret = 1;
 
-    gettimeofday(&h->last_send_time, &tz);
+    gettimeofday(&h->last_send_time, NULL);
     myseq = seqmap_add(h->i, h->num_sent, &h->last_send_time);
 
 #if defined(DEBUG) || defined(_DEBUG)
@@ -1671,11 +1715,11 @@ int send_ping(HOST_ENTRY* h)
 #endif /* DEBUG || _DEBUG */
 
     if (h->saddr.ss_family == AF_INET && socket4 >= 0) {
-        n = socket_sendto_ping_ipv4(socket4, (struct sockaddr*)&h->saddr, h->saddr_len, myseq, ident);
+        n = socket_sendto_ping_ipv4(socket4, (struct sockaddr*)&h->saddr, h->saddr_len, myseq, ident4);
     }
 #ifdef IPV6
     else if (h->saddr.ss_family == AF_INET6 && socket6 >= 0) {
-        n = socket_sendto_ping_ipv6(socket6, (struct sockaddr*)&h->saddr, h->saddr_len, myseq, ident);
+        n = socket_sendto_ping_ipv6(socket6, (struct sockaddr*)&h->saddr, h->saddr_len, myseq, ident6);
     }
 #endif
     else {
@@ -1827,19 +1871,22 @@ int decode_icmp_ipv4(
     unsigned short* id,
     unsigned short* seq)
 {
-    struct ip* ip = (struct ip*)reply_buf;
     struct icmp* icp;
     int hlen = 0;
 
+    if (!using_sock_dgram4) {
+        struct ip* ip = (struct ip*)reply_buf;
+
 #if defined(__alpha__) && __STDC__ && !defined(__GLIBC__)
-    /* The alpha headers are decidedly broken.
-     * Using an ANSI compiler, it provides ip_vhl instead of ip_hl and
-     * ip_v.  So, to get ip_hl, we mask off the bottom four bits.
-     */
-    hlen = (ip->ip_vhl & 0x0F) << 2;
+        /* The alpha headers are decidedly broken.
+         * Using an ANSI compiler, it provides ip_vhl instead of ip_hl and
+         * ip_v.  So, to get ip_hl, we mask off the bottom four bits.
+         */
+        hlen = (ip->ip_vhl & 0x0F) << 2;
 #else
-    hlen = ip->ip_hl << 2;
+        hlen = ip->ip_hl << 2;
 #endif
+    }
 
     if (reply_buf_len < hlen + ICMP_MINLEN) {
         /* too short */
@@ -1868,7 +1915,7 @@ int decode_icmp_ipv4(
 
         sent_icmp = (struct icmp*)(reply_buf + hlen + ICMP_MINLEN + sizeof(struct ip));
 
-        if (sent_icmp->icmp_type != ICMP_ECHO || ntohs(sent_icmp->icmp_id) != ident) {
+        if (sent_icmp->icmp_type != ICMP_ECHO || sent_icmp->icmp_id != ident4) {
             /* not caused by us */
             return 0;
         }
@@ -1917,7 +1964,7 @@ int decode_icmp_ipv4(
         return 0;
     }
 
-    *id = ntohs(icp->icmp_id);
+    *id = icp->icmp_id;
     *seq = ntohs(icp->icmp_seq);
 
     return 1;
@@ -1960,7 +2007,7 @@ int decode_icmp_ipv6(
 
         sent_icmp = (struct icmp6_hdr*)(reply_buf + sizeof(struct icmp6_hdr) + sizeof(struct ip));
 
-        if (sent_icmp->icmp6_type != ICMP_ECHO || ntohs(sent_icmp->icmp6_id) != ident) {
+        if (sent_icmp->icmp6_type != ICMP_ECHO || sent_icmp->icmp6_id != ident6) {
             /* not caused by us */
             return 0;
         }
@@ -2009,7 +2056,7 @@ int decode_icmp_ipv6(
         return 0;
     }
 
-    *id = ntohs(icp->icmp6_id);
+    *id = icp->icmp6_id;
     *seq = ntohs(icp->icmp6_seq);
 
     return 1;
@@ -2066,7 +2113,7 @@ int wait_for_reply(long wait_time)
         return 1;
     }
 
-    gettimeofday(&current_time, &tz);
+    gettimeofday(&current_time, NULL);
 
     /* Process ICMP packet and retrieve id/seq */
     if (response_addr.ss_family == AF_INET) {
@@ -2079,6 +2126,13 @@ int wait_for_reply(long wait_time)
                 &seq)) {
             return 1;
         }
+        if (id != ident4) {
+            return 1; /* packet received, but not the one we are looking for! */
+        }
+        if (using_sock_dgram4) {
+            /* IP header is not included in read SOCK_DGRAM ICMP responses */
+            result += sizeof(struct ip);
+        }
     }
 #ifdef IPV6
     else if (response_addr.ss_family == AF_INET6) {
@@ -2091,16 +2145,15 @@ int wait_for_reply(long wait_time)
                 &seq)) {
             return 1;
         }
+        if (id != ident6) {
+            return 1; /* packet received, but not the one we are looking for! */
+        }
     }
 #endif
     else {
         return 1;
     }
 
-    if (id != ident) {
-        return 1; /* packet received, but not the one we are looking for! */
-    }
-
     seqmap_value = seqmap_fetch(seq, &current_time);
     if (seqmap_value == NULL) {
         return 1;
@@ -2149,6 +2202,27 @@ int wait_for_reply(long wait_time)
         sum_replies += this_reply;
         h->total_time += this_reply;
         total_replies++;
+        
+        if (h->num_recv == 1) {
+            num_alive++;
+            if (verbose_flag || alive_flag) {
+                printf("%s", h->host);
+
+                if (verbose_flag)
+                    printf(" is alive");
+
+                if (elapsed_flag)
+                    printf(" (%s ms)", sprint_tm(this_reply));
+
+                if (addr_cmp((struct sockaddr*)&response_addr, (struct sockaddr*)&h->saddr)) {
+                    char buf[INET6_ADDRSTRLEN];
+                    getnameinfo((struct sockaddr*)&response_addr, sizeof(response_addr), buf, INET6_ADDRSTRLEN, NULL, 0, NI_NUMERICHOST);
+                    fprintf(stderr, " [<- %s]", buf);
+                }
+
+                printf("\n");
+            }
+        }
     }
 
     /* received ping is cool, so process it */
@@ -2185,27 +2259,6 @@ int wait_for_reply(long wait_time)
         }
     }
 
-    if (h->num_recv == 1) {
-        num_alive++;
-        if (verbose_flag || alive_flag) {
-            printf("%s", h->host);
-
-            if (verbose_flag)
-                printf(" is alive");
-
-            if (elapsed_flag)
-                printf(" (%s ms)", sprint_tm(this_reply));
-
-            if (addr_cmp((struct sockaddr*)&response_addr, (struct sockaddr*)&h->saddr)) {
-                char buf[INET6_ADDRSTRLEN];
-                getnameinfo((struct sockaddr*)&response_addr, sizeof(response_addr), buf, INET6_ADDRSTRLEN, NULL, 0, NI_NUMERICHOST);
-                fprintf(stderr, " [<- %s]", buf);
-            }
-
-            printf("\n");
-        }
-    }
-
     if (per_recv_flag) {
         if (timestamp_flag) {
             printf("[%lu.%06lu] ",
@@ -2240,7 +2293,6 @@ int wait_for_reply(long wait_time)
         remove_job(h);
     }
 
-    fflush(stdout);
     return num_jobs;
 }
 
@@ -2269,8 +2321,8 @@ void add_name(char* name)
     char addrbuf[256];
 
     /* getaddrinfo */
-    bzero(&hints, sizeof(struct addrinfo));
-    hints.ai_flags = 0;
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_flags = AI_UNUSABLE;
     hints.ai_socktype = SOCK_RAW;
     hints.ai_family = hints_ai_family;
     if (hints_ai_family == AF_INET) {
@@ -2336,7 +2388,7 @@ void add_name(char* name)
             }
 
             if (name_flag || rdns_flag) {
-                char nameaddrbuf[512];
+                char nameaddrbuf[512+3];
                 snprintf(nameaddrbuf, sizeof(nameaddrbuf) / sizeof(char), "%s (%s)", printname, addrbuf);
                 add_addr(name, nameaddrbuf, res->ai_addr, res->ai_addrlen);
             }
@@ -2580,7 +2632,7 @@ void timeval_add(struct timeval* a, long t_10u)
 
 char* sprint_tm(int t)
 {
-    static char buf[10];
+    static char buf[12];
 
     if (t < 0) {
         /* negative (unexpected) */
